/********************************************************************
  wraplib.c
    VIOSɸѴؿ stdlib åѡ
*********************************************************************/
#include <string.h>
#include <time.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <sys/wait.h>
#include "vios_lib.h"


void *
Calloc(size_t n, size_t size)
{
    void *ptr;

    if((ptr = calloc(n, size)) == NULL) errs_exit("calloc error");
    return(ptr);
}

void
Close(int fd)
{
    if(close(fd) == -1) errs_exit("close error");
}

void
Dup2(int fd1, int fd2)
{
    if(dup2(fd1, fd2) == -1) errs_exit("dup2 error");
}

int
Fcntl(int fd, int cmd, int arg)
{
    int	n;

    if((n = fcntl(fd, cmd, arg)) == -1) errs_exit("fcntl error");
    return(n);
}

void
Gettimeofday(struct timeval *tv, void *foo)
{
    if(gettimeofday(tv, foo) == -1) errs_exit("gettimeofday error");
    return;
}

int
Ioctl(int fd, int request, void *arg)
{
    int n;

    if((n = ioctl(fd, request, arg)) == -1) errs_exit("ioctl error");
    return n;
}

pid_t
Fork(void)
{
    pid_t pid;

    if((pid = fork()) == -1) errs_exit("fork error");
    return pid;
}

void *
Malloc(size_t size)
{
    void* ptr;

    if((ptr = malloc(size)) == NULL) errs_exit("malloc error");
    return ptr;
}

void 
Mktemp(char *temp)
{
    if((mktemp(temp) == NULL) || temp[0] == 0)
	err_exit("mktemp error");
}

void *
Mmap(void *addr, size_t len, int prot, int flags, int fd, off_t offset)
{
    void* ptr;

    if((ptr = mmap((char*)addr, len, prot, flags, fd, offset))
       == ((void *) -1))
	errs_exit("mmap error");
    return ptr;
}

int
Open(const char *pathname, int oflag, mode_t mode)
{
    int fd;

    if((fd = open(pathname, oflag, mode)) == -1)
	errs_exit("open error for %s", pathname);
    return fd;
}

void
Pipe(int *fds)
{
    if(pipe(fds) < 0) errs_exit("pipe error");
}

ssize_t
Read(int fd, void *ptr, size_t nbytes)
{
    ssize_t n;

    if((n = read(fd, ptr, nbytes)) == -1) errs_exit("read error");
    return n;
}

void
Sigaddset(sigset_t *set, int signo)
{
    if (sigaddset(set, signo) == -1) errs_exit("sigaddset error");
}

void
Sigdelset(sigset_t *set, int signo)
{
    if(sigdelset(set, signo) == -1) errs_exit("sigdelset error");
}

void
Sigemptyset(sigset_t *set)
{
    if(sigemptyset(set) == -1) errs_exit("sigemptyset error");
}

void
Sigfillset(sigset_t *set)
{
    if(sigfillset(set) == -1) errs_exit("sigfillset error");
}

int
Sigismember(const sigset_t *set, int signo)
{
    int n;

    if((n = sigismember(set, signo)) == -1)
	errs_exit("sigismember error");
    return n;
}

void
Sigpending(sigset_t *set)
{
    if(sigpending(set) == -1) errs_exit("sigpending error");
}

void
Sigprocmask(int how, const sigset_t *set, sigset_t *oset)
{
    if(sigprocmask(how, set, oset) == -1)
	errs_exit("sigprocmask error");
}

/* δ§äؿǤϤʤ */
char *
Strdup(const char *str)
{
    char* ptr;

    if((ptr = strdup(str)) == NULL) errs_exit("strdup error");
    return ptr;
}

long
Sysconf(int name)
{
    long val;

    errno = 0;	/* sysconf()ѹʤФΤޤ */
    if((val = sysconf(name)) == -1) errs_exit("sysconf error");
    return val;
}

#ifdef	HAVE_SYS_SYSCTL_H
void
Sysctl(int *name, u_int namelen, void *oldp, size_t *oldlenp,\
	   void *newp, size_t newlen)
{
    if(sysctl(name, namelen, oldp, oldlenp, newp, newlen) == -1)
	errs_exit("sysctl error");
}
#endif

void
Unlink(const char *pathname)
{
    if (unlink(pathname) == -1)
	errs_exit("unlink error for %s", pathname);
}

pid_t
Wait(int *iptr)
{
    pid_t pid;

    if((pid = wait(iptr)) == -1) errs_exit("wait error");
    return pid;
}

pid_t
Waitpid(pid_t pid, int *iptr, int options)
{
    pid_t retpid;

    if((retpid = waitpid(pid, iptr, options)) == -1)
	errs_exit("waitpid error");
    return retpid;
}

void
Write(int fd, void *ptr, size_t nbytes)
{
    if(write(fd, ptr, nbytes) != nbytes) errs_exit("write error");
}
