/*********************************************************************
  interpreter.cc
    VPE-i (main󥿡ץ꥿)¼¹
*********************************************************************/
#include <sched.h>
#include <math.h>
#include <pgm.h>
#include "vios_run.h"
#include "lang.h"
#include "om_table.h"
#include "om_table_sub.h"
#include "generator.h"
#include "interpreter.h"

Interpreter::Interpreter()
{
    sprintf(int_form, "%3d");
    sprintf(flt_form, "%1.5f");

    pc = -1;
    sp =  0;

    StepCode[JMP]	= &Interpreter::e_JMP;
    StepCode[JPF]	= &Interpreter::e_JPF;
    StepCode[LBL]	= &Interpreter::e_LBL;

    StepCode[GATH]	= &Interpreter::e_GATH;
    StepCode[iREAD]     = &Interpreter::e_iREAD;   //ɲ
    StepCode[fREAD]     = &Interpreter::e_fREAD;   //ɲ
    StepCode[iREADBIN]  = &Interpreter::e_iREADBIN;//ɲ
    StepCode[fREADBIN]  = &Interpreter::e_fREADBIN;//ɲ
    StepCode[SAVETXT]   = &Interpreter::e_SAVETXT; //ɲ
    StepCode[SAVEBIN]   = &Interpreter::e_SAVEBIN; //ɲ
    StepCode[LOAD]	= &Interpreter::e_LOAD;
    StepCode[SAVE]	= &Interpreter::e_SAVE;
    StepCode[iPUT]	= &Interpreter::e_iPUT;
    StepCode[fPUT]	= &Interpreter::e_fPUT;
    StepCode[sPUT]	= &Interpreter::e_sPUT;
    StepCode[RAND]	= &Interpreter::e_RAND;
    StepCode[SET]	= &Interpreter::e_SET;
    StepCode[WIDTH]	= &Interpreter::e_WIDTH;
    StepCode[HEIGH]	= &Interpreter::e_HEIGH;
    StepCode[DEPTH]	= &Interpreter::e_DEPTH;
    StepCode[MODULE]	= &Interpreter::e_MODULE;

    StepCode[CLS]	= &Interpreter::e_CLS;
    StepCode[REM]	= &Interpreter::e_REM;

    StepCode[REG0]	= &Interpreter::e_REG0;
    StepCode[REG1]	= &Interpreter::e_REG1;
    StepCode[REG2]	= &Interpreter::e_REG2;
    StepCode[REG3]	= &Interpreter::e_REG3;

    StepCode[iiPOP]	= &Interpreter::e_iiPOP;
    StepCode[ifPOP]	= &Interpreter::e_ifPOP;
    StepCode[fiPOP]	= &Interpreter::e_fiPOP;
    StepCode[ffPOP]	= &Interpreter::e_ffPOP;
    StepCode[ii1POP]	= &Interpreter::e_ii1POP;
    StepCode[if1POP]	= &Interpreter::e_if1POP;
    StepCode[fi1POP]	= &Interpreter::e_fi1POP;
    StepCode[ff1POP]	= &Interpreter::e_ff1POP;

    StepCode[iPUSH]	= &Interpreter::e_iPUSH;
    StepCode[fPUSH]	= &Interpreter::e_fPUSH;
    StepCode[sPUSH]	= &Interpreter::e_sPUSH;

    StepCode[iv1PUSH]	= &Interpreter::e_iv1PUSH;
    StepCode[fv1PUSH]	= &Interpreter::e_fv1PUSH;
    StepCode[ivPUSH]	= &Interpreter::e_ivPUSH;
    StepCode[fvPUSH]	= &Interpreter::e_fvPUSH;
    StepCode[svPUSH]	= &Interpreter::e_svPUSH;

    StepCode[ip1PUSH]	= &Interpreter::e_ip1PUSH;
    StepCode[fp1PUSH]	= &Interpreter::e_fp1PUSH;
    StepCode[ipPUSH]	= &Interpreter::e_ipPUSH;
    StepCode[fpPUSH]	= &Interpreter::e_fpPUSH;
    StepCode[spPUSH]	= &Interpreter::e_spPUSH;

    StepCode[ivXADDR]	= &Interpreter::e_ivXADDR;
    StepCode[fvXADDR]	= &Interpreter::e_fvXADDR;
    StepCode[ivYADDR]	= &Interpreter::e_ivYADDR;
    StepCode[fvYADDR]	= &Interpreter::e_fvYADDR;
    StepCode[ivZADDR]	= &Interpreter::e_ivZADDR;
    StepCode[fvZADDR]	= &Interpreter::e_fvZADDR;

    StepCode[ipXADDR]	= &Interpreter::e_ipXADDR;
    StepCode[fpXADDR]	= &Interpreter::e_fpXADDR;
    StepCode[ipYADDR]	= &Interpreter::e_ipYADDR;
    StepCode[fpYADDR]	= &Interpreter::e_fpYADDR;
    StepCode[ipZADDR]	= &Interpreter::e_ipZADDR;
    StepCode[fpZADDR]	= &Interpreter::e_fpZADDR;

    StepCode[iINT]	= &Interpreter::e_iINT;
    StepCode[iFLT]	= &Interpreter::e_iFLT;
    StepCode[iINC]	= &Interpreter::e_iINC;
    StepCode[iDEC]	= &Interpreter::e_iDEC;
    StepCode[iMIN]	= &Interpreter::e_iMIN;
    StepCode[iMUL]	= &Interpreter::e_iMUL;
    StepCode[iDIV]	= &Interpreter::e_iDIV;
    StepCode[iMOD]	= &Interpreter::e_iMOD;
    StepCode[iADD]	= &Interpreter::e_iADD;
    StepCode[iSUB]	= &Interpreter::e_iSUB;
    StepCode[iGTR]	= &Interpreter::e_iGTR;
    StepCode[iGEQ]	= &Interpreter::e_iGEQ;
    StepCode[iEQL]	= &Interpreter::e_iEQL;
    StepCode[iNEQ]	= &Interpreter::e_iNEQ;
    StepCode[iOR]	= &Interpreter::e_iOR;
    StepCode[iAND]	= &Interpreter::e_iAND;
    StepCode[iNOT]	= &Interpreter::e_iNOT;

    StepCode[fINT]	= &Interpreter::e_fINT;
    StepCode[fFLT]	= &Interpreter::e_fFLT;
    StepCode[fINC]	= &Interpreter::e_fINC;
    StepCode[fDEC]	= &Interpreter::e_fDEC;
    StepCode[fMIN]	= &Interpreter::e_fMIN;
    StepCode[fMUL]	= &Interpreter::e_fMUL;
    StepCode[fDIV]	= &Interpreter::e_fDIV;
    StepCode[fMOD]	= &Interpreter::e_fMOD;
    StepCode[fADD]	= &Interpreter::e_fADD;
    StepCode[fSUB]	= &Interpreter::e_fSUB;
    StepCode[fGTR]	= &Interpreter::e_fGTR;
    StepCode[fGEQ]	= &Interpreter::e_fGEQ;
    StepCode[fEQL]	= &Interpreter::e_fEQL;
    StepCode[fNEQ]	= &Interpreter::e_fNEQ;
    StepCode[fOR]	= &Interpreter::e_fOR;
    StepCode[fAND]	= &Interpreter::e_fAND;
    StepCode[fNOT]	= &Interpreter::e_fNOT;

}



// ΰμ¹ -------------------------------------------------
void
Interpreter::execute()
{
    // ٥ν --
    int step = 0;
    while(step <= pc) {
	if(code_area[step].op == LBL) {
	    label_area[code_area[step].sb.vl.ival] = step;
	}
	++step;
    }

    // ֥ɼ¹ --
    step = pc;
    pc   = 0;
    int flag = 0;

    while(pc <= step) {
//	fprintf(stderr,"pc:%d ...", pc);
	(this->*Interpreter::StepCode[code_area[pc].op])();
//	fprintf(stderr,"\n");
    }
//    fprintf(stderr,"------end------\n");
    //  ----------
    l_table.clear();
    pc   = -1;
}




// 饹åδ̿ ----------------------------------
void
Interpreter::e_JMP()
{
    pc = label_area[code_area[pc].sb.vl.ival];
}
void
Interpreter::e_JPF()
{
    if(stack_area[--sp].vl.ival == 0){
	pc = label_area[code_area[pc].sb.vl.ival];
    } else {
	++pc;
    }
}
void
Interpreter::e_LBL(){  ++pc;  }






// ؿ --------------------------------------
//
void
Interpreter::e_MODULE()
{
    module_tab.argc = stack_area[--sp].vl.ival - 1;

    for(int i=0; i<module_tab.argc; i++)
	module_tab.argv[module_tab.argc - i - 1] = stack_area[--sp].ag.dt;

    module_tab.name = stack_area[--sp].vl.sval;

    if(!om.conn) om.ConnectStart(1);

    timeval st, et;

    Gettimeofday(&st, NULL);
    om.doModule();
    Gettimeofday(&et, NULL);

    double t = et.tv_sec - st.tv_sec + \
	(et.tv_usec - st.tv_usec) * 0.000001;

    fprintf(stderr,"exec time: %1.4lf\n", t);
    ++pc;
}
void
Interpreter::e_SAVE()
{
    int	    argc = stack_area[--sp].vl.ival;
    vs_data* val = stack_area[--sp].ag.dt;
    char*   name = stack_area[--sp].vl.sval;

    // 顼ؤбȡХʥλǽ
    int* ival;
    switch(val->type) {
      case INT:
	Pgm_save(name, val->dim->width, 1,               (int*)val->val);
	break;
      case FLT:
	ival = new int[val->size];
	// ʤ
	for(int i=0; i<val->dim->width; i++)
	    ival[i] = (int)((FLOAT*)val->val)[i];
	Pgm_save(name, val->dim->width, 1,               ival);
	delete[] ival;
	break;
      case iIMG:
	Pgm_save(name, val->dim->width, val->dim->heigh, (int*)val->val);
	break;
      case fIMG:
	ival = new int[val->size];
	// ʤ
	for(int j=0; j<val->dim->heigh; j++)
	    for(int i=0; i<val->dim->width; i++) {
		int index = j * val->dim->width + i;
		ival[index] =  (int)((FLOAT*)val->val)[index];
	    }
	Pgm_save(name, val->dim->width, val->dim->heigh, ival);
	delete[] ival;
	break;
      default:
	fprintf(stderr,"imgSave: 3D data is not support.\n");
    }
    ++pc;
}


void //ɲ
Interpreter::e_iREAD(){//ɤ߹ߴؿο鼡Ƚ
  int x_size,y_size,z_size;
  int   argc   = stack_area[--sp].vl.ival;
  char* name   = stack_area[--sp].vl.sval;
  if(argc > 4){ // zʤ
    z_size = stack_area[--sp].vl.ival;
  }
  if(argc > 3){ // yʤ
    y_size = stack_area[--sp].vl.ival;
  }
  x_size = stack_area[--sp].vl.ival;
  char* path   = stack_area[--sp].vl.sval;
  if(argc == 3){
    g_table.regist1D(INT,name,x_size,path,1);
  } else if(argc == 4){
    g_table.regist2D(iIMG,name,x_size,y_size,path,1);
  } else {
    g_table.regist3D(iCUB,name,x_size,y_size,z_size,path,1);
  }
  ++pc;
}

void //ɲ
Interpreter::e_fREAD(){//ɤ߹ߴؿο鼡Ƚ
  int x_size,y_size,z_size;
  int   argc   = stack_area[--sp].vl.ival;
  char* name   = stack_area[--sp].vl.sval;
  if(argc > 4){ // zʤ
    z_size = stack_area[--sp].vl.ival;
  }
  if(argc > 3){ // yʤ
    y_size = stack_area[--sp].vl.ival;
  }
  x_size = stack_area[--sp].vl.ival;
  char* path   = stack_area[--sp].vl.sval;
  if(argc == 3){
    g_table.regist1D(FLT,name,x_size,path,1);
  } else if(argc == 4){
    g_table.regist2D(fIMG,name,x_size,y_size,path,1);
  } else {
    g_table.regist3D(fCUB,name,x_size,y_size,z_size,path,1);
  }
  ++pc;
}

void //ɲ
Interpreter::e_iREADBIN(){//ɤ߹ߴؿο鼡Ƚ
  int x_size,y_size,z_size;
  int   argc   = stack_area[--sp].vl.ival;
  char* name   = stack_area[--sp].vl.sval;
  if(argc > 4){ // zʤ
    z_size = stack_area[--sp].vl.ival;
  }
  if(argc > 3){ // yʤ
    y_size = stack_area[--sp].vl.ival;
  }
  x_size = stack_area[--sp].vl.ival;
  char* path   = stack_area[--sp].vl.sval;
  if(argc == 3){
    g_table.regist1D(INT,name,x_size,path,0);
  } else if(argc == 4){
    g_table.regist2D(iIMG,name,x_size,y_size,path,0);
  } else {
    g_table.regist3D(iCUB,name,x_size,y_size,z_size,path,0);
  }
  ++pc;
}

void //ɲ
Interpreter::e_fREADBIN(){//ɤ߹ߴؿο鼡Ƚ
  int x_size,y_size,z_size;
  int   argc   = stack_area[--sp].vl.ival;
  char* name   = stack_area[--sp].vl.sval;
  if(argc > 4){ // zʤ
    z_size = stack_area[--sp].vl.ival;
  }
  if(argc > 3){ // yʤ
    y_size = stack_area[--sp].vl.ival;
  }
  x_size = stack_area[--sp].vl.ival;
  char* path   = stack_area[--sp].vl.sval;
  if(argc == 3){
    g_table.regist1D(FLT,name,x_size,path,0);
  } else if(argc == 4){
    g_table.regist2D(fIMG,name,x_size,y_size,path,0);
  } else {
    g_table.regist3D(fCUB,name,x_size,y_size,z_size,path,0);
  }
  ++pc;
}

void //ɲ
Interpreter::e_SAVETXT(){ //¸ؿƥ
  int	  argc = stack_area[--sp].vl.ival;
  vs_data* val = stack_area[--sp].ag.dt;
  char*   path = stack_area[--sp].vl.sval;

  switch(val->type){
  case INT:
    iSave(path,val->dim->width,(int*)val->val);
    break;
  case FLT:
    fSave(path,val->dim->width,(FLOAT*)val->val);
    break;
  case iIMG:
    i2Save(path,val->dim->width,val->dim->heigh,(int*)val->val);
    break;
  case fIMG:
    f2Save(path,val->dim->width,val->dim->heigh,(FLOAT*)val->val);
    break; 
  case iCUB:
    i3Save(path,val->dim->width,val->dim->heigh,
	   val->dim->depth,(int*)val->val);
    break;
  case fCUB:
    f3Save(path,val->dim->width,val->dim->heigh,
	   val->dim->depth,(FLOAT*)val->val);
    break;
  }
  ++pc;
}

void //ɲ
Interpreter::e_SAVEBIN(){ //2¸ؿХʥ
  int	  argc = stack_area[--sp].vl.ival;
  vs_data* val = stack_area[--sp].ag.dt;
  char*   path = stack_area[--sp].vl.sval;
  switch(val->type){
  case INT:
    iSaveBin(path,val->size,(int*)val->val);
    break;
  case FLT:
    fSaveBin(path,val->size,(FLOAT*)val->val);
    break;
  case iIMG:
    iSaveBin(path,val->size,(int*)val->val);
    break;
  case fIMG:
    fSaveBin(path,val->size,(FLOAT*)val->val);
    break; 
  case iCUB:
    iSaveBin(path,val->size,(int*)val->val);
    break;
  case fCUB:
    fSaveBin(path,val->size,(FLOAT*)val->val);
    break;
  }
  ++pc;
}

void
Interpreter::e_LOAD()
{
    int	  argc = stack_area[--sp].vl.ival;
    char* name = stack_area[--sp].vl.sval;
    char* path = stack_area[--sp].vl.sval;

    g_table.regist(name, path);
    ++pc;
}

void
Interpreter::e_SET()
{
  int x, y, z, type;
  vs_data* var;
  
  switch(code_area[pc].sb.vl.ival) {
  case 5:
    z = stack_area[--sp].vl.ival;
  case 4:
    y = stack_area[--sp].vl.ival;
  case 3:
    x = stack_area[--sp].vl.ival;
    
    type	= stack_area[--sp].vl.ival;
    var	= stack_area[--sp].ag.dt;
    //fprintf(stderr,"type=%d,x=%d\n",type,x);
  }
  
  if(type == CACHE) {
    if(exec_info.host_num == 1){
      var->dim->cache = 0; // ۥȤοʤ饭åΰ̵̣ɲ
    } else {
      var->dim->cache = x;
    }
  } else if(type == dCOLCYC){ // åʬ䡣ɲ
    var->dtype = dCOLCYC;
    var->div_hint.x_divide = x; // COLʤΤxλ
    var->div_hint.y_divide = 0; // COLʤΤy0
  } else if(type == dROWCYC){
    var->dtype = dROWCYC;
    var->div_hint.x_divide = 0; // ROWʤΤx0
    var->div_hint.y_divide = x; // ROWʤΤyλ
  } else {
    if(x > 0) {
      switch(code_area[pc].sb.vl.ival) {
      case 5:
	var->div_hint.x_divide = x;
      case 4:
	var->div_hint.y_divide = y;
      case 3:
	var->div_hint.z_divide = z;
	var->dtype = dPIXEL;
      }
    } else {
      var->dtype	  = x; /* ƥޤ */
    }
  }
  ++pc;
}

void
Interpreter::e_iPUT()
{
    fprintf(stderr, "%3d", stack_area[--sp].vl.ival);
    ++pc;
}
void
Interpreter::e_fPUT()
{
    fprintf(stderr, "%1.2f", stack_area[--sp].vl.fval);
    ++pc;
}
void
Interpreter::e_sPUT()
{
    fprintf(stderr, stack_area[--sp].vl.sval);
    ++pc;
}
void
Interpreter::e_RAND()
{
    int	    argc = stack_area[--sp].vl.ival;
    stack_area[sp++].vl.ival = rand();
    ++pc;
}
void
Interpreter::e_WIDTH()
{
    int	    argc = stack_area[--sp].vl.ival;
    --sp;
    stack_area[sp].vl.ival = stack_area[sp].ag.dt->dim->width;
    ++sp; ++pc;
}
void
Interpreter::e_HEIGH()
{
    int	    argc = stack_area[--sp].vl.ival;
    --sp;
    stack_area[sp].vl.ival = stack_area[sp].ag.dt->dim->heigh;
    ++sp; ++pc;
}
void
Interpreter::e_DEPTH()
{
    int	    argc = stack_area[--sp].vl.ival;
    --sp;
    stack_area[sp].vl.ival = stack_area[sp].ag.dt->dim->depth;
    ++sp; ++pc;
}

void
Interpreter::e_GATH()
{
    int	    argc = stack_area[--sp].vl.ival;
    om.SendGather(stack_area[--sp].ag.dt);
    om.FromIP();
    ++pc;
}

void
Interpreter::e_REM()
{
    int	    argc = stack_area[--sp].vl.ival;
    int dv       = stack_area[--sp].vl.ival;
    vs_data* var = stack_area[--sp].ag.dt;

    om.SendGather(var);		// ޤ
    om.FromIP();

    var->dtype = dv;		// 
    g_table.remake(var);	// ꥻå
    ++pc;
}
void
Interpreter::e_CLS(){
    int	    argc = stack_area[--sp].vl.ival;
    ++pc; 
}



// ѿؤ -------------------------------------------------
#define _VS_VAL_POP(TYPE1, TYPE2, TYPE2_VAL)			\
    TYPE2 var  = stack_area[--sp].vl.TYPE2_VAL;			\
    int   addr = stack_area[--sp].vl.ival;			\
								\
    /* fprintf(stderr,"addr, var is (%d,%d)\n", addr, var); */  \
    *((TYPE1*)stack_area[--sp].ag.dt->val + addr) = var;	\
    ++pc

void
Interpreter::e_iiPOP(){  _VS_VAL_POP(int,   int, ival);  }
void
Interpreter::e_ifPOP(){  _VS_VAL_POP(FLOAT, int, ival);  }
void
Interpreter::e_fiPOP(){  _VS_VAL_POP(int, FLOAT, fval);  }
void
Interpreter::e_ffPOP(){  _VS_VAL_POP(FLOAT,FLOAT,fval);  }

#undef _VS_VAL_POP

#define _VS_VAL_1POP(TYPE1, TYPE2, TYPE2_VAL)		\
    TYPE2 var = stack_area[--sp].vl.TYPE2_VAL;		\
							\
    ((TYPE1*)stack_area[--sp].ag.dt->val)[0] = var;	\
    ++pc

void
Interpreter::e_ii1POP(){  _VS_VAL_1POP(int,   int, ival);  }
void
Interpreter::e_if1POP(){  _VS_VAL_1POP(FLOAT, int, ival);  }
void
Interpreter::e_fi1POP(){  _VS_VAL_1POP(int, FLOAT, fval);  }
void
Interpreter::e_ff1POP(){  _VS_VAL_1POP(FLOAT,FLOAT,fval);  }

#undef _VS_VAL_1POP

#define _VS_N_PUSH(VAL)						\
    stack_area[sp++].vl.VAL = code_area[pc].sb.ag.sb->dt.VAL;	\
    ++pc

void
Interpreter::e_iPUSH(){    _VS_N_PUSH(ival);    }
void
Interpreter::e_fPUSH(){    _VS_N_PUSH(fval);	}
void
Interpreter::e_sPUSH(){    _VS_N_PUSH(sval);	}

#undef _VS_N_PUSH

// ѿͤɤ
void
Interpreter::e_iv1PUSH()
{
    stack_area[sp++].vl.ival = *(  (int*)code_area[pc].sb.ag.dt->val);
    ++pc;
}
void
Interpreter::e_fv1PUSH()
{
    stack_area[sp++].vl.fval = *((FLOAT*)code_area[pc].sb.ag.dt->val);
    ++pc;
}
void
Interpreter::e_ivPUSH()
{
    int addr = stack_area[--sp].vl.ival;
    int* var = (int*)stack_area[--sp].ag.dt->val;

    stack_area[sp++].vl.ival = var[addr];
    ++pc;
}
void
Interpreter::e_fvPUSH()
{
    int addr   = stack_area[--sp].vl.ival;
    FLOAT* var = (FLOAT*)stack_area[--sp].ag.dt->val;

    stack_area[sp++].vl.fval = var[addr];
    ++pc;
}
void
Interpreter::e_svPUSH()
{
    stack_area[sp++].vl.sval = code_area[sp].sb.ag.sb->dt.sval;
    ++pc;
}

// ѿΥɥ쥹ɤʰĤǽʬ
#define _VS_LOAD_VAL					\
    stack_area[sp++].ag.dt = code_area[pc].sb.ag.dt;	\
    ++pc

void
Interpreter::e_ip1PUSH(){  _VS_LOAD_VAL;  }
void
Interpreter::e_fp1PUSH(){  _VS_LOAD_VAL;  }
void
Interpreter::e_ipPUSH() {  _VS_LOAD_VAL;  }
void
Interpreter::e_fpPUSH() {  _VS_LOAD_VAL;  }
void
Interpreter::e_spPUSH() {  _VS_LOAD_VAL;  }

#undef _VS_LOAD_VAL


void
Interpreter::e_ivXADDR()
{
    int addr = stack_area[--sp].vl.ival;
    int* var = (int*)stack_area[--sp].ag.dt->val;

    stack_area[sp++].vl.ival = var[addr];
    ++pc;
}
void
Interpreter::e_fvXADDR()
{
    int addr   = stack_area[--sp].vl.ival;
    FLOAT* var = (FLOAT*)stack_area[--sp].ag.dt->val;

    stack_area[sp++].vl.fval = var[addr];
    ++pc;
}

void
Interpreter::e_ivYADDR()
{
    int y = stack_area[--sp].vl.ival;
    int x = stack_area[--sp].vl.ival;
    vs_data* var = stack_area[--sp].ag.dt;

    stack_area[sp++].vl.ival = ((int*)var->val)[y * var->dim->width + x];
    ++pc;
}
void
Interpreter::e_fvYADDR()
{
    int y = stack_area[--sp].vl.ival;
    int x = stack_area[--sp].vl.ival;
    vs_data* var = stack_area[--sp].ag.dt;

    stack_area[sp++].vl.fval = ((FLOAT*)var->val)[y * var->dim->width + x];
    ++pc;
}

void
Interpreter::e_ivZADDR()
{
    int z = stack_area[--sp].vl.ival;
    int y = stack_area[--sp].vl.ival;
    int x = stack_area[--sp].vl.ival;
    vs_data* var = stack_area[--sp].ag.dt;

    stack_area[sp++].vl.ival = \
	((int*)var->val)[(z * var->dim->heigh + y) * var->dim->width + x];
    ++pc;
}
void
Interpreter::e_fvZADDR()
{
    int z = stack_area[--sp].vl.ival;
    int y = stack_area[--sp].vl.ival;
    int x = stack_area[--sp].vl.ival;
    vs_data* var = stack_area[--sp].ag.dt;

    stack_area[sp++].vl.fval = \
	((FLOAT*)var->val)[(z * var->dim->heigh + y) * var->dim->width + x];
    ++pc;
}


// ѿΥɥ쥹  ------------------------

void
Interpreter::e_ipXADDR(){  ++pc;  }
void
Interpreter::e_fpXADDR(){  ++pc;  }

#define _VS_YADDR_CALC							\
    int x, y;								\
    y = stack_area[--sp].vl.ival;					\
    x = stack_area[--sp].vl.ival;					\
    stack_area[sp].vl.ival = y * stack_area[sp-1].ag.dt->dim->width + x;\
    ++sp; ++pc

void
Interpreter::e_ipYADDR(){  _VS_YADDR_CALC;  }
void
Interpreter::e_fpYADDR(){  _VS_YADDR_CALC;  }

#undef _VS_YADDR_CALC

#define _VS_ZADDR_CALC							\
    int x, y, z;							\
    z = stack_area[--sp].vl.ival;					\
    y = stack_area[--sp].vl.ival;					\
    x = stack_area[--sp].vl.ival;					\
    DimInfo* info = stack_area[sp-1].ag.dt->dim;			\
    								        \
    stack_area[sp].vl.ival = (z * info->heigh + y) * info->width + x;	\
    ++sp; ++pc;

void
Interpreter::e_ipZADDR(){  _VS_ZADDR_CALC;  }
void
Interpreter::e_fpZADDR(){  _VS_ZADDR_CALC;  }

#undef _VS_ZADDR_CALC

void
Interpreter::e_iINT(){  ++pc;  }
void
Interpreter::e_iFLT()
{
    --sp;
    stack_area[sp].vl.fval = (int)stack_area[sp].vl.ival;
    ++sp; ++pc;
}
void
Interpreter::e_iINC()
{
    stack_area[sp-1].vl.ival++;
    ++pc;
}
void
Interpreter::e_iDEC()
{
    stack_area[sp-1].vl.ival--;
    ++pc;
}
void
Interpreter::e_iMIN()
{
    --sp;
    stack_area[sp].vl.ival = - stack_area[sp].vl.ival;
    ++sp; ++pc;
}

void
Interpreter::e_fINT()
{
    --sp;
    stack_area[sp].vl.ival = (int)stack_area[sp].vl.fval;
    ++sp; ++pc;
}
void
Interpreter::e_fFLT(){  ++pc;  }
void
Interpreter::e_fINC()
{
    stack_area[sp-1].vl.fval++;
    ++pc;
}
void
Interpreter::e_fDEC()
{
    stack_area[sp-1].vl.fval--;
    ++pc;
}
void
Interpreter::e_fMIN()
{
    --sp;
    stack_area[sp].vl.fval = - stack_area[sp].vl.fval;
    ++sp; ++pc;
}

/****************************************************************
  ߡint, FLOATκä黻θƤʤȤդ
 ****************************************************************/

#define	_VS_INT_CALC(OP)						\
    --sp; --sp;								\
    stack_area[sp].vl.ival =						\
                (stack_area[sp].vl.ival OP stack_area[sp+1].vl.ival);	\
    ++sp; ++pc

void
Interpreter::e_iMUL(){	_VS_INT_CALC(*);  }
void
Interpreter::e_iDIV(){	_VS_INT_CALC(/);  }
void
Interpreter::e_iSUB(){	_VS_INT_CALC(-);  }
void
Interpreter::e_iADD(){  _VS_INT_CALC(+);  }
void
Interpreter::e_iMOD(){	_VS_INT_CALC(%);  }
void
Interpreter::e_iGTR(){	_VS_INT_CALC(>);  }
void
Interpreter::e_iGEQ(){	_VS_INT_CALC(>=); }
void
Interpreter::e_iEQL(){	_VS_INT_CALC(==); }
void
Interpreter::e_iNEQ(){	_VS_INT_CALC(!=); }
void
Interpreter::e_iOR() {	_VS_INT_CALC(||); }
void
Interpreter::e_iAND(){	_VS_INT_CALC(&&); }

#undef _VS_INT_CALC

void
Interpreter::e_iNOT()
{
    --sp;
    stack_area[sp].vl.ival = !(stack_area[sp].vl.ival);
    ++sp; ++pc;
}

#define	_VS_FLT_CALC(OP)						\
    --sp; --sp;								\
    stack_area[sp].vl.fval =						\
                (stack_area[sp].vl.fval OP stack_area[sp+1].vl.fval);	\
    ++sp; ++pc

void
Interpreter::e_fMUL(){  _VS_FLT_CALC(*);  }
void
Interpreter::e_fDIV(){  _VS_FLT_CALC(/);  }
void
Interpreter::e_fMOD()
{
    --sp; --sp;
    stack_area[sp].vl.fval =	\
                fmod(stack_area[sp].vl.fval, stack_area[sp+1].vl.fval);
    ++sp; ++pc;
}
void
Interpreter::e_fADD(){  _VS_FLT_CALC(+);  }
void
Interpreter::e_fSUB(){  _VS_FLT_CALC(-);  }
void
Interpreter::e_fGTR(){  _VS_FLT_CALC(>);  }
void
Interpreter::e_fGEQ(){  _VS_FLT_CALC(>=);  }
void
Interpreter::e_fEQL(){  _VS_FLT_CALC(==);  }
void
Interpreter::e_fNEQ(){  _VS_FLT_CALC(!=);  }
void
Interpreter::e_fOR() {  _VS_FLT_CALC(||);  }
void
Interpreter::e_fAND(){  _VS_FLT_CALC(&&);  }

#undef _VS_FLT_CALC

void
Interpreter::e_fNOT()
{
    --sp;
    stack_area[sp].vl.ival = !(stack_area[sp].vl.fval);
    ++sp; ++pc;
}
